/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     |
    \\  /    A nd           | Copyright (C) 1991-2009 OpenCFD Ltd.
     \\/     M anipulation  |
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the
    Free Software Foundation; either version 2 of the License, or (at your
    option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM; if not, write to the Free Software Foundation,
    Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA

\*---------------------------------------------------------------------------*/

#include "SSG.H"
#include "addToRunTimeSelectionTable.H"

#include "backwardsCompatibilityWallFunctions.H"
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

namespace Foam
{
namespace incompressible
{
namespace RASModels
{

// * * * * * * * * * * * * * * Static Data Members * * * * * * * * * * * * * //

defineTypeNameAndDebug(SSG, 0);
addToRunTimeSelectionTable(RASModel, SSG, dictionary);

// * * * * * * * * * * * * * * * * Constructors  * * * * * * * * * * * * * * //

SSG::SSG
(
    const volVectorField& U,
    const surfaceScalarField& phi,
    transportModel& lamTransportModel
)
:
    RASModel(typeName, U, phi, lamTransportModel),

    C1_
    (
        dimensioned<scalar>::lookupOrAddToDict
        (
            "C1",
            coeffDict_,
            1.44
        )
    ),
    C2_
    (
        dimensioned<scalar>::lookupOrAddToDict
        (
            "C2",
            coeffDict_,
            1.92
        )
    ),
    Cmu_
    (
        dimensioned<scalar>::lookupOrAddToDict
        (
            "Cmu",
            coeffDict_,
            0.09
        )
    ),
    Cs_
    (
        dimensioned<scalar>::lookupOrAddToDict
        (
            "Cs",
            coeffDict_,
            0.25
        )
    ),
    Ceps_
    (
        dimensioned<scalar>::lookupOrAddToDict
        (
            "Ceps",
            coeffDict_,
            0.15
        )
    ),
    sigmaR_
    (
        dimensioned<scalar>::lookupOrAddToDict
        (
            "sigmaR",
            coeffDict_,
            0.81967
        )
    ),
    sigmaEps_
    (
        dimensioned<scalar>::lookupOrAddToDict
        (
            "sigmaEps",
            coeffDict_,
            1.3
        )
    ),
    Cg1_
    (
        dimensioned<scalar>::lookupOrAddToDict
        (
            "Cg1",
            coeffDict_,
            3.4
        )
    ),

    Cg1s_
    (
        dimensioned<scalar>::lookupOrAddToDict
        (
            "Cg1s",
            coeffDict_,
            1.8
        )
    ),

    Cg2_
    (
        dimensioned<scalar>::lookupOrAddToDict
        (
            "Cg2",
            coeffDict_,
            4.2
        )
    ),
    Cg3_
    (
        dimensioned<scalar>::lookupOrAddToDict
        (
            "Cg3",
            coeffDict_,
            0.8
        )
    ),

    Cg3s_
    (
        dimensioned<scalar>::lookupOrAddToDict
        (
            "Cg3s",
            coeffDict_,
            1.3
        )
    ),

    Cg4_
    (
        dimensioned<scalar>::lookupOrAddToDict
        (
            "Cg4",
            coeffDict_,
            1.25
        )
    ),

    Cg5_
    (
        dimensioned<scalar>::lookupOrAddToDict
        (
            "Cg5",
            coeffDict_,
            0.4
        )
    ),
    R_
    (
        IOobject
        (
            "R",
            runTime_.timeName(),
            mesh_,
            IOobject::NO_READ,
            IOobject::AUTO_WRITE
        ),
        autoCreateR("R", mesh_)
    ),
    k_
    (
        IOobject
        (
            "k",
            runTime_.timeName(),
            mesh_,
            IOobject::NO_READ,
            IOobject::AUTO_WRITE
        ),
        autoCreateK("k", mesh_)
    ),
    epsilon_
    (
        IOobject
        (
            "epsilon",
            runTime_.timeName(),
            mesh_,
            IOobject::NO_READ,
            IOobject::AUTO_WRITE
        ),
        autoCreateEpsilon("epsilon", mesh_)
    ),
    nut_
    (
        IOobject
        (
            "nut",
            runTime_.timeName(),
            mesh_,
            IOobject::NO_READ,
            IOobject::AUTO_WRITE
        ),
        autoCreateNut("nut", mesh_)
    )

{
    nut_ = Cmu_*sqr(k_)/(epsilon_ + epsilonSmall_);
    nut_.correctBoundaryConditions();

    printCoeffs();
}


// * * * * * * * * * * * * * * * Member Functions  * * * * * * * * * * * * * //

template<class Cmpt, template<class> class PatchField, class GeoMesh>
tmp<GeometricField<Tensor<Cmpt>, PatchField, GeoMesh> >
SSG::symm2full
(
    GeometricField<SymmTensor<Cmpt>, PatchField, GeoMesh>& symm
)
{
    typedef GeometricField<Tensor<Cmpt>, PatchField, GeoMesh> FieldType;
    tmp<FieldType> tFull
    (
        new FieldType
        (
           IOobject
           (
               symm.name() + "FullTensor",
               symm.mesh().time().timeName(),
               symm.mesh(),
               IOobject::NO_READ,
               IOobject::NO_WRITE
           ),
           symm.mesh(),
           dimensionedTensor("0", symm.dimensions(), Tensor<Cmpt>::zero),
           symm.boundaryField().types()
        )
    );

    FieldType& full = tFull();

    // manipulate components
    full.replace(Tensor<Cmpt>::XX, symm.component(SymmTensor<Cmpt>::XX));
    full.replace(Tensor<Cmpt>::YY, symm.component(SymmTensor<Cmpt>::YY));
    full.replace(Tensor<Cmpt>::ZZ, symm.component(SymmTensor<Cmpt>::ZZ));
    full.replace(Tensor<Cmpt>::XY, symm.component(SymmTensor<Cmpt>::XY));
    full.replace(Tensor<Cmpt>::YZ, symm.component(SymmTensor<Cmpt>::YZ));
    full.replace(Tensor<Cmpt>::XZ, symm.component(SymmTensor<Cmpt>::XZ));
    full.replace(Tensor<Cmpt>::ZX, symm.component(SymmTensor<Cmpt>::XZ));
    full.replace(Tensor<Cmpt>::YX, symm.component(SymmTensor<Cmpt>::XY));
    full.replace(Tensor<Cmpt>::ZY, symm.component(SymmTensor<Cmpt>::YZ));

    return tFull;
}

tmp<volSymmTensorField> SSG::devReff() const
{
    return tmp<volSymmTensorField>
    (
        new volSymmTensorField
        (
            IOobject
            (
                "devRhoReff",
                runTime_.timeName(),
                mesh_,
                IOobject::NO_READ,
                IOobject::NO_WRITE
            ),
            R_ - nu()*dev(twoSymm(fvc::grad(U_)))
        )
    );
}


tmp<fvVectorMatrix> SSG::divDevReff(volVectorField& U) const
{
        return
        (
            fvc::div(R_)
          + fvc::laplacian(nut_, U, "laplacian(nuEff,U)")
          - fvm::laplacian(nuEff(), U)
        );
}

bool SSG::read()
{
    if (RASModel::read())
    {
        Cg1_.readIfPresent(coeffDict());
        Cg1s_.readIfPresent(coeffDict());
        Cg2_.readIfPresent(coeffDict());
        Cg3_.readIfPresent(coeffDict());
        Cg3s_.readIfPresent(coeffDict());
        Cg4_.readIfPresent(coeffDict());
        Cg5_.readIfPresent(coeffDict());
        C1_.readIfPresent(coeffDict());
        C2_.readIfPresent(coeffDict());
        Cmu_.readIfPresent(coeffDict());
        Cs_.readIfPresent(coeffDict());
        Ceps_.readIfPresent(coeffDict());

        sigmaEps_.readIfPresent(coeffDict());
        sigmaR_.readIfPresent(coeffDict());

        return true;
    }
    else
    {
        return false;
    }
}

void SSG::correct()
{
    if (mesh_.changing())
    {
        R_.correctBoundaryConditions();
        bound(epsilon_, epsilon0_);
    }

    RASModel::correct();

    if (!turbulence_)
    {
        return;
    }

    volSymmTensorField P = -twoSymm(R_ & fvc::grad(U_));
    volScalarField G("RASModel::G", 0.5*mag(tr(P)));

    // Update epsilon and G at the wall
    epsilon_.boundaryField().updateCoeffs();

    // Dissipation equation
    tmp<fvScalarMatrix> epsEqn
    (
        fvm::ddt(epsilon_)
      + fvm::div(phi_, epsilon_)
      + fvm::SuSp(-fvc::div(phi_), epsilon_)
    //- fvm::laplacian(Ceps*(k_/epsilon_)*R_, epsilon_)
      - fvm::laplacian(DepsilonEff(), epsilon_)
     ==
        C1_*G*epsilon_/k_
      - fvm::Sp(C2_*epsilon_/k_, epsilon_)
    );

    epsEqn().relax();

    epsEqn().boundaryManipulate(epsilon_.boundaryField());

    solve(epsEqn);
    bound(epsilon_, epsilon0_);

    // Reynolds stress equation
    const fvPatchList& patches = mesh_.boundary();

    forAll(patches, patchi)
    {
        const fvPatch& curPatch = patches[patchi];

        //if (curPatch.isWall())
	if (isA<wallFvPatch>(curPatch))
        {
            forAll(curPatch, facei)
            {
                label faceCelli = curPatch.faceCells()[facei];
                P[faceCelli] *=
                    min(G[faceCelli]/(0.5*mag(tr(P[faceCelli])) + SMALL), 1.0);
            }
        }
    }

        volSymmTensorField bij("bij", 0.5*dev(R_/k_));
        volTensorField gradU = fvc::grad(U_);
        volSymmTensorField Sij("Sij", dev(symm(gradU)));
        volTensorField Wij = skew(gradU);

    tmp<fvSymmTensorMatrix> REqn
        (
            fvm::ddt(R_)
            + fvm::div(phi_, R_)
            + fvm::SuSp(-fvc::div(phi_), R_)
          //- fvm::laplacian(Cs*(k_/epsilon_)*R_, R_)
            - fvm::laplacian(DREff(), R_)
            + fvm::Sp(epsilon_/k_, R_)
            ==
            P                                         // production tensor
            - (Cg1_*epsilon_ + Cg1s_*G)*bij
            + Cg2_*epsilon_*dev(symm(bij & bij))
            + (Cg3_ - Cg3s_*sqrt(bij && bij))*k_*Sij
            + Cg4_*k_*dev(twoSymm(symm2full(bij) & Sij))
            + Cg5_*k_*(twoSymm(bij & Wij))
        );

    REqn().relax();
    solve(REqn);

    R_.max
    (
        dimensionedSymmTensor
        (
            "zero",
            R_.dimensions(),
            symmTensor
            (
                k0_.value(), -GREAT, -GREAT,
                             k0_.value(), -GREAT,
                                          k0_.value()
            )
        )
    );

    k_ == 0.5*tr(R_);
    bound(k_, k0_);


    // Re-calculate turbulent viscosity
    nut_ = Cmu_*sqr(k_)/epsilon_;
    nut_.correctBoundaryConditions();

    // Correct wall shear stresses

    forAll(patches, patchi)
    {
        const fvPatch& curPatch = patches[patchi];

        //if (curPatch.isWall())
	if (isA<wallFvPatch>(curPatch))
        {
            symmTensorField& Rw = R_.boundaryField()[patchi];

            const scalarField& nutw = nut_.boundaryField()[patchi];

            vectorField snGradU = U_.boundaryField()[patchi].snGrad();

            const vectorField& faceAreas
                = mesh_.Sf().boundaryField()[patchi];

            const scalarField& magFaceAreas
                = mesh_.magSf().boundaryField()[patchi];

            forAll(curPatch, facei)
            {
                // Calculate near-wall velocity gradient
                tensor gradUw
                    = (faceAreas[facei]/magFaceAreas[facei])*snGradU[facei];

                // Calculate near-wall shear-stress tensor
                tensor tauw = -nutw[facei]*2*symm(gradUw);

                // Reset the shear components of the stress tensor
                Rw[facei].xy() = tauw.xy();
                Rw[facei].xz() = tauw.xz();
                Rw[facei].yz() = tauw.yz();
            }
        }
    }
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

} // End namespace RASModels
} // End namespace incompressible
} // End namespace Foam

// ************************************************************************* //
