/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     |
    \\  /    A nd           | Copyright (C) 1991-2009 OpenCFD Ltd.
     \\/     M anipulation  |
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the
    Free Software Foundation; either version 2 of the License, or (at your
    option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM; if not, write to the Free Software Foundation,
    Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA

Class
    Foam::incompressible::RASModels::SSG

Description
    SSG Model.

SourceFiles
    SSG.C

\*---------------------------------------------------------------------------*/

#ifndef SSG_H
#define SSG_H

#include "RASModel.H"

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

namespace Foam
{
namespace incompressible
{
namespace RASModels
{

/*---------------------------------------------------------------------------*\
                            Class SSG Declaration
\*---------------------------------------------------------------------------*/

class SSG
:
    public RASModel
{
// private data

        // Model coefficients
            //const fvMesh& mesh_;

            dimensionedScalar C1_;
            dimensionedScalar C2_;
            dimensionedScalar Cmu_;
            dimensionedScalar Cs_;
            dimensionedScalar Ceps_;

            dimensionedScalar sigmaR_;
            dimensionedScalar sigmaEps_;

            dimensionedScalar Cg1_;
            dimensionedScalar Cg1s_;
            dimensionedScalar Cg2_;
            dimensionedScalar Cg3_;
            dimensionedScalar Cg3s_;
            dimensionedScalar Cg4_;
            dimensionedScalar Cg5_;

       // Fields

            volSymmTensorField R_;
            volScalarField k_;
            volScalarField epsilon_;
            volScalarField nut_;

    // Protected member functions
protected:

        //- Construct a tensor field out of a symmTensorField
        // The dot product between two SymmTensor is not correctly defined in
        // SymmTensorI.H.
        template<class Cmpt, template<class> class PatchField, class GeoMesh>
        static tmp<GeometricField<Tensor<Cmpt>, PatchField, GeoMesh> >
        symm2full
        (
            GeometricField<SymmTensor<Cmpt>, PatchField, GeoMesh>& symm
        );

public:

    //- Runtime type information
    TypeName("SSG");

    // Constructors

        //- Construct from components
        SSG
        (
            const volVectorField& U,
            const surfaceScalarField& phi,
            transportModel& transport
        );

    //- Destructor
    virtual ~SSG()
    {}


    // Member Functions

        //- Return the turbulence viscosity
        virtual tmp<volScalarField> nut() const
        {
           // return CmuKE_*sqr(k_)/(epsilon_ + epsilonSmall_);
            return nut_;
        }

        //- Return the effective diffusivity for R
        tmp<volScalarField> DREff() const
        {
            return tmp<volScalarField>
            (
                new volScalarField("DREff", nut_/sigmaR_ + nu())
            );
        }

        //- Return the effective diffusivity for epsilon
        tmp<volScalarField> DepsilonEff() const
        {
            return tmp<volScalarField>
            (
                new volScalarField("DepsilonEff", nut_/sigmaEps_ + nu())
            );
        }

        //- Return the turbulence kinetic energy
        virtual tmp<volScalarField> k() const
        {
            return k_;
        }

        //- Return the turbulence kinetic energy dissipation rate
        virtual tmp<volScalarField> epsilon() const
        {
            return epsilon_;
        }

        //- Return the Reynolds stress tensor
        virtual tmp<volSymmTensorField> R() const
        {
            return R_;
        }

        //- Return the effective stress tensor including the laminar stress
        virtual tmp<volSymmTensorField> devReff() const;

        //- Return the source term for the momentum equation
        virtual tmp<fvVectorMatrix> divDevReff(volVectorField& U) const;

        //- Solve the turbulence equations and correct the turbulence viscosity
        virtual void correct();

        //- Read RASProperties dictionary
        virtual bool read();
};


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

} // End namespace RASModels
} // End namespace incompressible
} // End namespace Foam

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

#endif

// ************************************************************************* //
